// SPDX-License-Identifier: MIT

pub static APP_ID: &str = "org.startup_disk.StartupDisk";
pub static APP_VERSION: &str = env!("CARGO_PKG_VERSION");
pub static RESOURCE_BASE: &str = "/org/startup-disk/StartupDisk";
